import collections
from collections.abc import Sequence
from typing import Any

import pystac

from stac_generator.core.base import (
    CollectionGenerator,
    ItemGenerator,
    StacCollectionConfig,
)
from stac_generator.core.base.schema import SourceConfig
from stac_generator.core.base.utils import read_source_config
from stac_generator.core.point import PointGenerator
from stac_generator.core.point.schema import PointConfig
from stac_generator.core.raster import RasterGenerator
from stac_generator.core.raster.schema import RasterConfig
from stac_generator.core.vector import VectorGenerator
from stac_generator.core.vector.schema import VectorConfig

EXTENSION_MAP: dict[str, type[SourceConfig]] = {
    "csv": PointConfig,
    "txt": PointConfig,
    "geotiff": RasterConfig,
    "tiff": RasterConfig,
    "tif": RasterConfig,
    "zip": VectorConfig,
    "geojson": VectorConfig,
    "json": VectorConfig,
    "gpkg": VectorConfig,  # Can also contain raster data. TODO: overhaul interface
    "shp": VectorConfig,
}

CONFIG_GENERATOR_MAP: dict[type[SourceConfig], type[ItemGenerator]] = {
    VectorConfig: VectorGenerator,
    RasterConfig: RasterGenerator,
    PointConfig: PointGenerator,
}

BaseConfig_T = (
    str
    | SourceConfig
    | dict[str, Any]
    | Sequence[str]
    | Sequence[SourceConfig]
    | Sequence[dict[str, Any]]
)
Config_T = BaseConfig_T | Sequence[BaseConfig_T]


class StacGeneratorFactory:
    @staticmethod
    def extract_item_config(item: pystac.Item) -> SourceConfig:
        if "stac_generator" not in item.properties:
            raise ValueError(f"Missing stac_generator properties for item: {item.id}")
        ext = item.properties["stac_generator"]["location"].split(".")[-1]
        handler = StacGeneratorFactory.get_config_handler(ext)
        return handler.model_validate(item.properties["stac_generator"])

    @staticmethod
    def match_handler(  # noqa: C901
        configs: Config_T,
    ) -> list[ItemGenerator]:
        # Read in configs
        _configs: list[SourceConfig] = []

        def handle_dict_config(config: dict[str, Any]) -> None:
            if "id" not in config:
                raise ValueError("Missing id in a config item.")
            if "location" not in config:
                raise ValueError(f"Missing location in a config item: {config['id']}")
            ext = config["location"].split(".")[-1]
            config_handler = StacGeneratorFactory.get_config_handler(ext)
            _configs.append(config_handler(**config))

        def handle_str_config(config: str) -> None:
            config_dicts = read_source_config(config)
            for item in config_dicts:
                handle_dict_config(item)

        def handle_source_config(config: SourceConfig) -> None:
            _configs.append(config)

        def handle_base_config(config: BaseConfig_T) -> None:
            if isinstance(config, str):
                handle_str_config(config)
            elif isinstance(config, SourceConfig):
                handle_source_config(config)
            elif isinstance(config, dict):
                handle_dict_config(config)
            else:
                raise TypeError(f"Invalid config item type: {type(config)}")

        def handle_config(config: Config_T) -> None:
            if isinstance(config, str | dict | SourceConfig):
                handle_base_config(config)
            elif hasattr(config, "__len__"):
                for item in config:
                    handle_config(item)
            else:
                raise TypeError(f"Invalid config type: {type(config)}")

        handle_config(configs)

        handler_map: dict[type[ItemGenerator], list[SourceConfig]] = collections.defaultdict(list)
        for config in _configs:
            generator_handler = StacGeneratorFactory.get_generator_handler(config)
            handler_map[generator_handler].append(config)

        generators = []
        for k, v in handler_map.items():
            generators.append(k(v))
        return generators

    @staticmethod
    def register_config_handler(
        extension: str, handler: type[SourceConfig], force: bool = False
    ) -> None:
        if extension in EXTENSION_MAP and not force:
            raise ValueError(
                f"Handler for extension: {extension} already exists: {EXTENSION_MAP[extension].__name__}. If this is intentional, use register_config_handler with force=True"
            )
        if not issubclass(handler, SourceConfig):
            raise ValueError("Registered handler must be an instance of a subclass of SourceConfig")
        EXTENSION_MAP[extension] = handler

    @staticmethod
    def register_generator_handler(
        config: SourceConfig, handler: type[ItemGenerator], force: bool = False
    ) -> None:
        config_type = type(config)
        if config_type in CONFIG_GENERATOR_MAP and not force:
            raise ValueError(
                f"Handler for config: {config_type.__name__} already exists: {CONFIG_GENERATOR_MAP[config_type].__name__}. If this is intentional, use register_generator_handler with force=True"
            )
        if not issubclass(handler, ItemGenerator):
            raise ValueError(
                "Registered handler must be an instance of a subclass of ItemGenerator"
            )
        CONFIG_GENERATOR_MAP[config_type] = handler

    @staticmethod
    def get_config_handler(extension: str) -> type[SourceConfig]:
        """Factory method to get SourceConfig class based on given extension

        :param extension: file extension
        :type extension: str
        :raises ValueError: if SourceConfig handler class for this file extension has not been registered_
        :return: handler class
        :rtype: type[SourceConfig]
        """
        if extension not in EXTENSION_MAP:
            raise ValueError(
                f"No SourceConfig matches extension: {extension}. Either change the extension or register a handler with the method `register_config_handler`"
            )
        return EXTENSION_MAP[extension]

    @staticmethod
    def get_generator_handler(config: SourceConfig) -> type[ItemGenerator]:
        config_type = type(config)
        if config_type not in CONFIG_GENERATOR_MAP:
            raise ValueError(
                f"No ItemGenerator for config of type: {config_type.__name__}. To register a handler, use the method StacGeneratorFactor.register_generator_handler"
            )
        return CONFIG_GENERATOR_MAP[config_type]

    @staticmethod
    def get_stac_generator(
        source_configs: Config_T, collection_config: StacCollectionConfig
    ) -> CollectionGenerator:
        handlers = StacGeneratorFactory.match_handler(source_configs)
        return CollectionGenerator(collection_config, handlers)



import logging
from collections.abc import Sequence

import geopandas as gpd
import pystac

from stac_generator._types import CsvMediaType
from stac_generator.core.base.generator import VectorGenerator
from stac_generator.core.base.schema import ASSET_KEY, ColumnInfo
from stac_generator.core.base.utils import _read_csv
from stac_generator.core.point.schema import PointConfig

logger = logging.getLogger(__name__)


def read_csv(
    src_path: str,
    X_coord: str,
    Y_coord: str,
    epsg: int,
    Z_coord: str | None = None,
    T_coord: str | None = None,
    date_format: str = "ISO8601",
    columns: Sequence[str] | Sequence[ColumnInfo] | None = None,
) -> gpd.GeoDataFrame:
    """Read in csv from local disk

    Users must provide at the bare minimum the location of the csv, and the names of the columns to be
    treated as the X and Y coordinates. By default, will read in all columns in the csv. If columns and groupby
    columns are provided, will selectively read specified columns together with the coordinate columns (X, Y, T).

    :param src_path: path to csv file
    :type src_path: str
    :param X_coord: name of X field
    :type X_coord: str
    :param Y_coord: name of Y field
    :type Y_coord: str
    :param epsg: epsg code
    :type epsg: int
    :param Z_coord: name of Z field
    :type Z_coord: str
    :param T_coord: name of time field, defaults to None
    :type T_coord: str | None, optional
    :param date_format: format to pass to pandas to parse datetime, defaults to "ISO8601"
    :type date_format: str, optional
    :param columns: band information, defaults to None
    :type columns: list[str] | list[ColumnInfo] | None, optional
    :return: read dataframe
    :rtype: pd.DataFrame
    """
    df = _read_csv(
        src_path=src_path,
        required=[X_coord, Y_coord],
        optional=[Z_coord] if Z_coord else None,
        date_col=T_coord,
        date_format=date_format,
        columns=columns,
    )

    return gpd.GeoDataFrame(df, geometry=gpd.points_from_xy(df[X_coord], df[Y_coord], crs=epsg))


class PointGenerator(VectorGenerator[PointConfig]):
    """ItemGenerator class that handles point data in csv format"""

    def create_item_from_config(self, source_config: PointConfig) -> pystac.Item:
        """Create item from source csv config

        :param source_config: config which contains csv metadata
        :type source_config: PointConfig
        :return: stac metadata of the item described in source_config
        :rtype: pystac.Item
        """
        assets = {
            ASSET_KEY: pystac.Asset(
                href=source_config.location,
                description="Raw csv data",
                roles=["data"],
                media_type=CsvMediaType,
            )
        }
        raw_df = read_csv(
            source_config.location,
            source_config.X,
            source_config.Y,
            source_config.epsg,
            source_config.Z,
            source_config.T,
            source_config.date_format,
            source_config.column_info,
        )
        if source_config.T is not None:
            start_datetime = raw_df[source_config.T].min()
            end_datetime = raw_df[source_config.T].max()
        else:
            start_datetime, end_datetime = None, None

        properties = source_config.to_properties()
        return self.df_to_item(
            raw_df,
            assets,
            source_config,
            properties={"stac_generator": properties},
            epsg=source_config.epsg,
            start_datetime=start_datetime,
            end_datetime=end_datetime,
        )



import logging
import re
from typing import cast

import geopandas as gpd
import pystac
from pyproj.crs.crs import CRS

from stac_generator.core.base.generator import VectorGenerator as BaseVectorGenerator
from stac_generator.core.base.schema import ASSET_KEY
from stac_generator.core.base.utils import _read_csv
from stac_generator.core.vector.schema import VectorConfig

logger = logging.getLogger(__name__)


def extract_epsg(crs: CRS) -> tuple[int, bool]:
    """Extract epsg information from crs object.
    If epsg info can be extracted directly from crs, return that value.
    Otherwise, try to convert the crs info to WKT2 and extract EPSG using regex

    Note that this method may yield unreliable result

    :param crs: crs object
    :type crs: CRS
    :return: epsg information
    :rtype: tuple[int, bool] - epsg code and reliability flag
    """
    if (result := crs.to_epsg()) is not None:
        return (result, True)
    # Handle WKT1 edge case
    wkt = crs.to_wkt()
    match = re.search(r'ID\["EPSG",(\d+)\]', wkt)
    if match:
        return (int(match.group(1)), True)
    # No match - defaults to 4326
    logger.warning(
        "Cannot determine epsg from vector file. Either provide it in the config or change the source file. Defaults to 4326 but can be incorrect."
    )
    return (4326, False)


class VectorGenerator(BaseVectorGenerator[VectorConfig]):
    """ItemGenerator class that handles vector data with common vector formats - i.e (shp, zipped shp, gpkg, geojson)"""

    def create_item_from_config(self, source_config: VectorConfig) -> pystac.Item:
        """Create item from vector config

        :param source_config: config information
        :type source_config: VectorConfig
        :raises ValueError: if config epsg information is different from epsg information from vector file
        :return: stac metadata of the file described by source_config
        :rtype: pystac.Item
        """
        assets = {
            ASSET_KEY: pystac.Asset(
                href=str(source_config.location),
                media_type=pystac.MediaType.GEOJSON
                if source_config.location.endswith(".geojson")
                else "application/x-shapefile",
                roles=["data"],
                description="Raw vector data",
            )
        }
        logger.debug(f"Reading file from {source_config.location}")

        # Only read relevant fields
        columns = [
            col["name"] if isinstance(col, dict) else col for col in source_config.column_info
        ]
        # Throw exceptions if column_info contains invalid column
        raw_df = gpd.read_file(source_config.location, layer=source_config.layer)

        if columns and not set(columns).issubset(set(raw_df.columns)):
            raise ValueError(f"Invalid columns: {set(columns) - set(raw_df.columns)}")

        # Validate EPSG user-input vs extracted
        epsg, _ = extract_epsg(raw_df.crs)

        start_datetime, end_datetime = None, None
        # Read join file
        if source_config.join_file:
            if source_config.join_attribute_vector not in raw_df.columns:
                raise ValueError(
                    f"If a join file is provided, expects join attribute vector: {source_config.join_attribute_vector} to be a valid attribute of the vector file."
                )
            # Try reading join file and raise errors if columns not provided
            try:
                join_df = _read_csv(
                    src_path=source_config.join_file,
                    required=[cast(str, source_config.join_field)],
                    date_format=source_config.date_format,
                    date_col=source_config.join_T_column,
                    columns=source_config.join_column_info,
                )
            except ValueError as e:
                raise ValueError(
                    f"Join file associated with vector file: {source_config.id} may not have the specified column"
                ) from e
            if source_config.join_T_column:
                start_datetime = join_df[source_config.join_T_column].min()
                end_datetime = join_df[source_config.join_T_column].max()

        # Make properties
        properties = source_config.to_properties()

        return self.df_to_item(
            raw_df,
            assets,
            source_config,
            properties={"stac_generator": properties},
            epsg=epsg,
            start_datetime=start_datetime,
            end_datetime=end_datetime,
        )



from __future__ import annotations

import json
from typing import cast

import pystac
import rasterio
from pyproj import CRS
from pyproj.transformer import Transformer
from pystac.extensions.eo import Band, EOExtension
from pystac.extensions.projection import ItemProjectionExtension, ProjectionExtension
from pystac.extensions.raster import AssetRasterExtension, DataType, RasterBand
from shapely import box, to_geojson

from stac_generator.core.base.generator import ItemGenerator
from stac_generator.core.base.schema import ASSET_KEY
from stac_generator.core.base.utils import calculate_timezone

from .schema import RasterConfig


class RasterGenerator(ItemGenerator[RasterConfig]):
    """Raster Generator"""

    def create_item_from_config(self, source_config: RasterConfig) -> pystac.Item:
        """Generate Raster Item from config

        :param source_config: raster config - must contain band info
        :type source_config: RasterConfig
        :raises ValueError: if epsg code is provided in config but does not match that extracted from the asset
        :return: generated item
        :rtype: pystac.Item
        """
        with rasterio.open(source_config.location) as src:
            bounds = src.bounds
            transform = src.transform
            crs = cast(CRS, src.crs)
            shape = list(src.shape)

            # Convert to 4326 for bbox and geometry
            transformer = Transformer.from_crs(crs, 4326, always_xy=True)
            minx, miny = transformer.transform(bounds.left, bounds.bottom)
            maxx, maxy = transformer.transform(bounds.right, bounds.top)
            bbox: tuple[float, float, float, float] = (minx, miny, maxx, maxy)

            # Create geometry as Shapely Polygon
            geometry = box(*bbox)
            geometry_geojson = json.loads(to_geojson(geometry))

            # Process datetime
            item_tz = calculate_timezone(geometry)
            item_ts = source_config.get_datetime(item_tz)

            # Get EPSG
            epsg = crs.to_epsg()

            # Create STAC Item
            # Start datetime and end_datetime are set to be collection datetime for Raster data
            item = pystac.Item(
                id=source_config.id,
                geometry=geometry_geojson,
                bbox=list(bbox),
                datetime=item_ts,
                properties={"stac_generator": source_config.to_properties()},
                start_datetime=item_ts,
                end_datetime=item_ts,
            )

            # Projection extension
            proj_ext = ItemProjectionExtension.ext(item, add_if_missing=True)
            affine_transform = [
                rasterio.transform.from_bounds(*bbox, shape[1], shape[0])[i] for i in range(9)
            ]
            proj_ext.apply(epsg=epsg, wkt2=crs.to_wkt(), shape=shape, transform=affine_transform)

            # Initialize extensions
            EOExtension.ext(item, add_if_missing=True)
            proj_ext = ProjectionExtension.ext(item, add_if_missing=True)
            proj_ext.apply(epsg=crs.to_epsg(), shape=shape, transform=list(transform)[:9])

            # Create EO and Raster bands
            eo_bands = []
            raster_bands = []
            for band_info in source_config.band_info:
                eo_band = Band.create(
                    name=band_info["name"].lower(),
                    common_name=band_info.get("common_name", None),
                    center_wavelength=band_info.get("wavelength", None),
                    description=band_info.get("description", None),
                )
                eo_bands.append(eo_band)

                raster_band = RasterBand.create(
                    nodata=band_info.get("nodata", 0),
                    data_type=DataType(band_info.get("data_type"))
                    if band_info.get("data_type", None)
                    else None,
                )
                raster_bands.append(raster_band)

            # Create Asset and Add to Item
            asset = pystac.Asset(
                href=source_config.location,
                media_type=pystac.MediaType.GEOTIFF,
                roles=["data"],
                title="Raster Data",
            )
            item.add_asset(ASSET_KEY, asset)

            # Apply Raster Extension to the Asset
            asset_raster_ext = AssetRasterExtension.ext(asset, add_if_missing=True)
            asset_raster_ext.apply(bands=raster_bands)

            # Add eo:bands to Asset
            # Apply EO bands to the asset using AssetEOExtension
            asset_eo_ext = EOExtension.ext(asset, add_if_missing=True)
            asset_eo_ext.apply(bands=eo_bands)

            return item



import json
from pathlib import Path

import pytest

from stac_generator.core.base.generator import CollectionGenerator
from stac_generator.core.base.schema import StacCollectionConfig
from stac_generator.core.base.utils import read_source_config
from stac_generator.core.vector.generator import VectorGenerator
from stac_generator.core.vector.schema import VectorConfig

CONFIG_JSON = Path("tests/files/integration_tests/vector/config/vector_config.json")

CONFIG_CSV = Path("tests/files/integration_tests/vector/config/vector_config.csv")

GENERATED_DIR = Path("tests/files/integration_tests/vector/generated")


JSON_CONFIGS = read_source_config(str(CONFIG_JSON))
CSV_CONFIGS = read_source_config(str(CONFIG_CSV))
ITEM_IDS = [item["id"] for item in JSON_CONFIGS]


@pytest.fixture(scope="module")
def json_vector_generator() -> VectorGenerator:
    return VectorGenerator(JSON_CONFIGS)


@pytest.fixture(scope="module")
def vector_generator() -> VectorGenerator:
    return VectorGenerator(CSV_CONFIGS)


@pytest.fixture(scope="module")
def collection_generator(vector_generator: VectorGenerator) -> CollectionGenerator:
    return CollectionGenerator(StacCollectionConfig(id="collection"), generators=[vector_generator])


@pytest.mark.parametrize("item_idx", range(len(JSON_CONFIGS)), ids=ITEM_IDS)
def test_generator_given_item_expects_matched_generated_item(
    item_idx: int, json_vector_generator: VectorGenerator
) -> None:
    config = JSON_CONFIGS[item_idx]
    expected_path = GENERATED_DIR / f"{config['id']}/{config['id']}.json"
    with expected_path.open() as file:
        expected = json.load(file)
    actual = json_vector_generator.create_item_from_config(VectorConfig(**config)).to_dict()
    assert expected["id"] == actual["id"]
    assert expected["bbox"] == actual["bbox"]
    assert expected["properties"] == actual["properties"]
    assert expected["assets"] == actual["assets"]
    assert expected["geometry"] == actual["geometry"]


@pytest.mark.parametrize("item_idx", range(len(CSV_CONFIGS)), ids=ITEM_IDS)
def test_generator_given_item_expects_matched_generated_item_csv_config_version(
    item_idx: int, vector_generator: VectorGenerator
) -> None:
    config = JSON_CONFIGS[item_idx]
    expected_path = GENERATED_DIR / f"{config['id']}/{config['id']}.json"
    with expected_path.open() as file:
        expected = json.load(file)
    actual = vector_generator.create_item_from_config(VectorConfig(**config)).to_dict()
    assert expected["id"] == actual["id"]
    assert expected["bbox"] == actual["bbox"]
    assert expected["properties"] == actual["properties"]
    assert expected["assets"] == actual["assets"]
    assert expected["geometry"] == actual["geometry"]


def test_collection_generator(collection_generator: CollectionGenerator) -> None:
    actual = collection_generator.create_collection().to_dict()
    expected_path = GENERATED_DIR / "collection.json"
    with expected_path.open() as file:
        expected = json.load(file)
    assert actual["extent"] == expected["extent"]



import json
from pathlib import Path

import pytest

from stac_generator import StacGeneratorFactory
from stac_generator.core.base import StacCollectionConfig
from stac_generator.core.base.generator import CollectionGenerator

FILE_PATH = Path("tests/files/integration_tests")
GENERATED_PATH = FILE_PATH / "composite/generated"
COLLECTION_ID = "collection"
collection_config = StacCollectionConfig(id=COLLECTION_ID)
CONFIGS_LIST = [
    str(FILE_PATH / "point/config/point_config.csv"),
    str(FILE_PATH / "vector/config/vector_config.json"),
    str(FILE_PATH / "raster/config/raster_config.csv"),
]
COMPOSITE_CONFIG = FILE_PATH / "composite/config/composite_config.json"


# One single config that is a composition of multiple configs
composite_generator = StacGeneratorFactory.get_stac_generator(
    [str(COMPOSITE_CONFIG)], collection_config
)
# Config provided as a list of configs
list_generator = StacGeneratorFactory.get_stac_generator(CONFIGS_LIST, collection_config)


@pytest.mark.parametrize(
    "generator",
    (composite_generator, list_generator),
    ids=["Composite Config", "List Configs"],
)
def test_generator_factory(
    generator: CollectionGenerator,
) -> None:
    collection = generator.create_collection()
    expected_collection_path = GENERATED_PATH / "collection.json"
    with expected_collection_path.open() as file:
        expected_collection = json.load(file)
    actual_collection = collection.to_dict()
    assert actual_collection["extent"] == expected_collection["extent"]
    for item in collection.get_items(recursive=True):
        config_loc = GENERATED_PATH / f"{item.id}/{item.id}.json"
        with config_loc.open("r") as file:
            expected = json.load(file)
        actual = item.to_dict()
        assert expected["id"] == actual["id"]
        assert expected["bbox"] == actual["bbox"]
        assert expected["geometry"] == actual["geometry"]
        assert expected["properties"] == actual["properties"]
        assert expected["assets"] == actual["assets"]



import json
import logging
import urllib.parse
from collections.abc import Sequence
from pathlib import Path
from typing import Any, cast

import httpx
import numpy as np
import pandas as pd
import yaml
from shapely import Geometry, centroid
from timezonefinder import TimezoneFinder

from stac_generator.core.base.schema import ColumnInfo

SUPPORTED_URI_SCHEMES = ["http", "https"]
logger = logging.getLogger(__name__)

TZFinder = TimezoneFinder()


def parse_href(base_url: str, collection_id: str, item_id: str | None = None) -> str:
    """Generate href for collection or item based on id

    :param base_url: path to catalog.
    :type base_url: str
    :param collection_id: collection id
    :type collection_id: str
    :param item_id: item id, defaults to None
    :type item_id: str | None, optional
    :return: uri to collection or item
    :rtype: str
    """
    if item_id:
        return urllib.parse.urljoin(base_url, f"{collection_id}/{item_id}")
    return urllib.parse.urljoin(base_url, f"{collection_id}")


def href_is_stac_api_endpoint(href: str) -> bool:
    """Check if href points to a resource behind a stac api

    :param href: path to resource
    :type href: str
    :return: local or non-local
    :rtype: bool
    """
    output = urllib.parse.urlsplit(href)
    return output.scheme == ""


def force_write_to_stac_api(url: str, id: str, json: dict[str, Any]) -> None:
    """Force write a json object to a stac api endpoint.
    This function will try sending a POST request and if a 409 error is encountered,
    try sending a PUT request. Other exceptions if occured will be raised

    :param url: path to stac resource for creation/update
    :type url: str
    :param json: json object
    :type json: dict[str, Any]
    """
    try:
        logger.debug(f"sending POST request to {url}")
        response = httpx.post(url=url, json=json)
        response.raise_for_status()
    except httpx.HTTPStatusError as err:
        if err.response.status_code == 409:
            logger.debug(f"sending PUT request to {url}")
            response = httpx.put(url=f"{url}/{id}", json=json)
            response.raise_for_status()
        else:
            raise err


def read_source_config(href: str) -> list[dict[str, Any]]:
    logger.debug(f"reading config file from {href}")
    if not href.endswith(("json", "yaml", "yml", "csv")):
        raise ValueError(
            "Unsupported extension. We currently allow json/yaml/csv files to be used as config"
        )
    if href.endswith(".csv"):
        df = pd.read_csv(href)
        df.replace(np.nan, None, inplace=True)
        return cast(list[dict[str, Any]], df.to_dict("records"))
    if not href.startswith(("http", "https")):
        with Path(href).open("r") as file:
            if href.endswith(("yaml", "yml")):
                result = yaml.safe_load(file)
            if href.endswith("json"):
                result = json.load(file)
    else:
        response = httpx.get(href, follow_redirects=True)
        response.raise_for_status()
        if href.endswith("json"):
            result = response.json()
        if href.endswith(("yaml", "yml")):
            result = yaml.safe_load(response.content.decode("utf-8"))

    if isinstance(result, dict):
        return [result]
    if isinstance(result, list):
        return result
    raise ValueError(f"Expects config to be read as a list of dictionary. Provided: {type(result)}")


def calculate_timezone(geometry: Geometry) -> str:
    """Calculate timezone from geometry

    :param geometry: geometry object in EPSG:4326 crs
    :type geometry: Geometry
    :raises ValueError: very exceptional cases where tzfinder cannot determine timezone
    :return: timezone string
    :rtype: str
    """
    point = centroid(geometry)
    # Use TimezoneFinder to get the timezone
    timezone_str = TZFinder.timezone_at(lng=point.x, lat=point.y)

    if not timezone_str:
        raise ValueError(
            f"Could not determine timezone for coordinates: lon={point.x}, lat={point.y}"
        )
    return timezone_str


def _read_csv(
    src_path: str,
    required: Sequence[str] | None = None,
    optional: Sequence[str] | None = None,
    date_col: str | None = None,
    date_format: str | None = "ISO8601",
    columns: Sequence[str] | Sequence[ColumnInfo] | None = None,
) -> pd.DataFrame:
    logger.debug(f"reading csv from path: {src_path}")
    parse_dates: list[str] | bool = [date_col] if isinstance(date_col, str) else False
    usecols: list[str] | None = None
    # If band info is provided, only read in the required columns + the X and Y coordinates
    if columns:
        usecols = [item["name"] if isinstance(item, dict) else item for item in columns]
        if required:
            usecols.extend(required)
        if optional:
            usecols.extend(optional)
        if date_col:
            usecols.append(date_col)
    return pd.read_csv(
        filepath_or_buffer=src_path,
        usecols=usecols,
        date_format=date_format,
        parse_dates=parse_dates,
    )
